"use strict";
/**
 * Copyright (c) 2015-present, Waysact Pty Ltd
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SubresourceIntegrityPlugin = void 0;
const crypto_1 = require("crypto");
const path_1 = require("path");
const fs_1 = require("fs");
const assert = __importStar(require("typed-assert"));
const thisPluginName = "webpack-subresource-integrity";
// https://www.w3.org/TR/2016/REC-SRI-20160623/#cryptographic-hash-functions
const standardHashFuncNames = ["sha256", "sha384", "sha512"];
let getHtmlWebpackPluginHooks = null;
function getTagSrc(tag) {
    if (!["script", "link"].includes(tag.tagName) || !tag.attributes) {
        return undefined;
    }
    if (typeof tag.attributes.href === "string") {
        return tag.attributes.href;
    }
    if (typeof tag.attributes.src === "string") {
        return tag.attributes.src;
    }
    return undefined;
}
const normalizePath = (p) => p.replace(/\?.*$/, "").split(path_1.sep).join("/");
const placeholderPrefix = "*-*-*-CHUNK-SRI-HASH-";
const computeIntegrity = (hashFuncNames, source) => {
    const result = hashFuncNames
        .map((hashFuncName) => hashFuncName +
        "-" +
        crypto_1.createHash(hashFuncName)
            .update(typeof source === "string" ? Buffer.from(source, "utf-8") : source)
            .digest("base64"))
        .join(" ");
    return result;
};
const makePlaceholder = (hashFuncNames, id) => {
    const placeholder = `${placeholderPrefix}${id}`;
    const filler = computeIntegrity(hashFuncNames, placeholder);
    return placeholderPrefix + filler.substring(placeholderPrefix.length);
};
function findChunks(chunk) {
    const allChunks = new Set();
    const groupsVisited = new Set();
    function addIfNotExist(set, item) {
        if (set.has(item))
            return true;
        set.add(item);
        return false;
    }
    (function recurseChunk(childChunk) {
        function recurseGroup(group) {
            if (addIfNotExist(groupsVisited, group.id))
                return;
            group.chunks.forEach(recurseChunk);
            group.childrenIterable.forEach(recurseGroup);
        }
        if (addIfNotExist(allChunks, childChunk))
            return;
        Array.from(childChunk.groupsIterable).forEach(recurseGroup);
    })(chunk);
    return allChunks;
}
/**
 * The webpack-subresource-integrity plugin.
 *
 * @public
 */
class SubresourceIntegrityPlugin {
    /**
     * Create a new instance.
     *
     * @public
     */
    constructor(options = {}) {
        /**
         * @internal
         */
        this.emittedMessages = new Set();
        /**
         * @internal
         */
        this.assetIntegrity = new Map();
        /**
         * @internal
         */
        this.inverseAssetIntegrity = new Map();
        /**
         * @internal
         */
        this.hwpPublicPath = null;
        /**
         * @internal
         */
        this.addMissingIntegrityHashes = (assets) => {
            Object.keys(assets).forEach((assetKey) => {
                const asset = assets[assetKey];
                let source;
                try {
                    source = asset.source();
                }
                catch (_) {
                    return;
                }
                this.updateAssetIntegrity(assetKey, computeIntegrity(this.options.hashFuncNames, source));
            });
        };
        /**
         * @internal
         */
        this.replaceAsset = (compiler, assets, hashByChunkId, chunkFile) => {
            const oldSource = assets[chunkFile].source();
            const hashFuncNames = this.options.hashFuncNames;
            const newAsset = new compiler.webpack.sources.ReplaceSource(assets[chunkFile], chunkFile);
            Array.from(hashByChunkId.entries()).forEach((idAndHash) => {
                const magicMarker = makePlaceholder(hashFuncNames, idAndHash[0]);
                const magicMarkerPos = oldSource.indexOf(magicMarker);
                if (magicMarkerPos >= 0) {
                    newAsset.replace(magicMarkerPos, magicMarkerPos + magicMarker.length - 1, idAndHash[1], chunkFile);
                }
            });
            assets[chunkFile] = newAsset;
            return newAsset;
        };
        /**
         * @internal
         */
        this.processChunk = (chunk, compilation, assets) => {
            const hashByChunkId = new Map();
            Array.from(findChunks(chunk))
                .reverse()
                .forEach((childChunk) => {
                const files = Array.from(childChunk.files);
                let sourcePath = files[files.length - 1];
                if (!sourcePath) {
                    return;
                }
                if (assets[sourcePath]) {
                    this.warnIfHotUpdate(compilation, assets[sourcePath].source());
                    const newAsset = this.replaceAsset(compilation.compiler, assets, hashByChunkId, sourcePath);
                    const integrity = computeIntegrity(this.options.hashFuncNames, newAsset.source());
                    if (childChunk.id !== null) {
                        hashByChunkId.set(childChunk.id, integrity);
                    }
                    this.updateAssetIntegrity(sourcePath, integrity);
                    compilation.updateAsset(sourcePath, (x) => x, (assetInfo) => assetInfo && {
                        ...assetInfo,
                        contenthash: Array.isArray(assetInfo.contenthash)
                            ? [...new Set([...assetInfo.contenthash, integrity])]
                            : assetInfo.contenthash
                                ? [assetInfo.contenthash, integrity]
                                : integrity,
                    });
                }
                else {
                    this.warnOnce(compilation, `No asset found for source path '${sourcePath}', options are ${Object.keys(assets).join(", ")}`);
                }
            });
        };
        /**
         * @internal
         */
        this.addAttribute = (compilation, elName, source) => {
            if (!compilation.outputOptions.crossOriginLoading) {
                this.errorOnce(compilation, "webpack option output.crossOriginLoading not set, code splitting will not work!");
            }
            return compilation.compiler.webpack.Template.asString([
                source,
                elName + ".integrity = __webpack_require__.sriHashes[chunkId];",
                elName +
                    ".crossOrigin = " +
                    JSON.stringify(compilation.outputOptions.crossOriginLoading) +
                    ";",
            ]);
        };
        /**
         * @internal
         */
        this.processAssets = (compilation, assets) => {
            Array.from(compilation.chunks)
                .filter((chunk) => chunk.hasRuntime())
                .forEach((chunk) => {
                this.processChunk(chunk, compilation, assets);
            });
            this.addMissingIntegrityHashes(assets);
        };
        /**
         * @internal
         */
        this.hwpAssetPath = (src) => {
            assert.isNotNull(this.hwpPublicPath);
            return path_1.relative(this.hwpPublicPath, src);
        };
        /**
         * @internal
         */
        this.getIntegrityChecksumForAsset = (assets, src) => {
            if (this.assetIntegrity.has(src)) {
                return this.assetIntegrity.get(src);
            }
            const normalizedSrc = normalizePath(src);
            const normalizedKey = Object.keys(assets).find((assetKey) => normalizePath(assetKey) === normalizedSrc);
            if (normalizedKey) {
                return this.assetIntegrity.get(normalizedKey);
            }
            return undefined;
        };
        /**
         * @internal
         */
        this.processTag = (compilation, tag) => {
            if (tag.attributes &&
                Object.prototype.hasOwnProperty.call(tag.attributes, "integrity")) {
                return;
            }
            const tagSrc = getTagSrc(tag);
            if (!tagSrc) {
                return;
            }
            const src = this.hwpAssetPath(tagSrc);
            let integrity = this.getIntegrityChecksumForAsset(compilation.assets, src);
            if (!integrity) {
                const candidate = path_1.join(compilation.compiler.outputPath, src);
                try {
                    integrity = computeIntegrity(this.options.hashFuncNames, fs_1.readFileSync(candidate));
                }
                catch (e) {
                    if (e.code !== "ENOENT") {
                        throw e;
                    }
                }
            }
            if (!integrity) {
                this.errorOnce(compilation, "Could not determine integrity for asset at path " + src);
                return;
            }
            tag.attributes.integrity = integrity;
            tag.attributes.crossorigin =
                compilation.compiler.options.output.crossOriginLoading || "anonymous";
        };
        /**
         * @internal
         */
        this.setup = (compilation) => {
            if (!this.validateOptions(compilation) || !this.isEnabled(compilation)) {
                return;
            }
            if (typeof compilation.outputOptions.chunkLoading === "string" &&
                ["require", "async-node"].includes(compilation.outputOptions.chunkLoading)) {
                this.warnOnce(compilation, "This plugin is not useful for non-web targets.");
                return;
            }
            compilation.hooks.processAssets.tap({
                name: thisPluginName,
                stage: compilation.compiler.webpack.Compilation
                    .PROCESS_ASSETS_STAGE_OPTIMIZE_INLINE,
            }, (records) => {
                return this.processAssets(compilation, records);
            });
            compilation.hooks.afterProcessAssets.tap(thisPluginName, (records) => {
                Object.keys(records).forEach((assetName) => {
                    if (records[assetName].source().includes(placeholderPrefix) &&
                        !assetName.endsWith(".map") // FIXME
                    ) {
                        this.errorOnce(compilation, `Asset ${assetName} contains unresolved integrity placeholders`);
                    }
                });
            });
            compilation.compiler.webpack.optimize.RealContentHashPlugin.getCompilationHooks(compilation).updateHash.tap(thisPluginName, (input, oldHash) => {
                const assetKey = this.inverseAssetIntegrity.get(oldHash);
                if (assetKey && input.length === 1) {
                    const newIntegrity = computeIntegrity(this.options.hashFuncNames, input[0]);
                    this.inverseAssetIntegrity.delete(oldHash);
                    this.assetIntegrity.delete(assetKey);
                    this.updateAssetIntegrity(assetKey, newIntegrity);
                    return newIntegrity;
                }
                // FIXME: remove type hack pending https://github.com/webpack/webpack/pull/12642#issuecomment-784744910
                return undefined;
            });
            if (getHtmlWebpackPluginHooks) {
                getHtmlWebpackPluginHooks(compilation).beforeAssetTagGeneration.tapPromise(thisPluginName, async (pluginArgs) => {
                    this.hwpPublicPath = pluginArgs.assets.publicPath;
                    ["js", "css"].forEach((fileType) => {
                        pluginArgs.assets[fileType + "Integrity"] = pluginArgs.assets[fileType].map((filePath) => this.getIntegrityChecksumForAsset(compilation.assets, this.hwpAssetPath(filePath)));
                    });
                    return pluginArgs;
                });
                getHtmlWebpackPluginHooks(compilation).alterAssetTagGroups.tapPromise({
                    name: thisPluginName,
                    stage: 10000,
                }, async (data) => {
                    this.addMissingIntegrityHashes(compilation.assets);
                    data.headTags
                        .concat(data.bodyTags)
                        .forEach((tag) => this.processTag(compilation, tag));
                    return data;
                });
            }
            const { mainTemplate } = compilation;
            mainTemplate.hooks.jsonpScript.tap(thisPluginName, (source) => this.addAttribute(compilation, "script", source));
            mainTemplate.hooks.linkPreload.tap(thisPluginName, (source) => this.addAttribute(compilation, "link", source));
            mainTemplate.hooks.localVars.tap(thisPluginName, (source, chunk) => {
                const allChunks = findChunks(chunk);
                const includedChunks = chunk.getChunkMaps(false).hash;
                if (Object.keys(includedChunks).length > 0) {
                    return compilation.compiler.webpack.Template.asString([
                        source,
                        "__webpack_require__.sriHashes = " +
                            JSON.stringify(Array.from(allChunks).reduce((sriHashes, depChunk) => {
                                if (depChunk.id !== null &&
                                    includedChunks[depChunk.id.toString()]) {
                                    sriHashes[depChunk.id] = makePlaceholder(this.options.hashFuncNames, depChunk.id);
                                }
                                return sriHashes;
                            }, {})) +
                            ";",
                    ]);
                }
                return source;
            });
        };
        /**
         * @internal
         */
        this.validateOptions = (compilation) => {
            if (this.isEnabled(compilation) &&
                !compilation.compiler.options.output.crossOriginLoading) {
                this.warnOnce(compilation, 'SRI requires a cross-origin policy, defaulting to "anonymous". ' +
                    "Set webpack option output.crossOriginLoading to a value other than false " +
                    "to make this warning go away. " +
                    "See https://w3c.github.io/webappsec-subresource-integrity/#cross-origin-data-leakage");
            }
            return this.validateHashFuncNames(compilation);
        };
        /**
         * @internal
         */
        this.validateHashFuncNames = (compilation) => {
            if (!Array.isArray(this.options.hashFuncNames)) {
                this.error(compilation, "options.hashFuncNames must be an array of hash function names, " +
                    "instead got '" +
                    this.options.hashFuncNames +
                    "'.");
                return false;
            }
            else if (this.options.hashFuncNames.length === 0) {
                this.error(compilation, "Must specify at least one hash function name.");
                return false;
            }
            else if (!this.options.hashFuncNames.every(this.validateHashFuncName.bind(this, compilation))) {
                return false;
            }
            else {
                this.warnStandardHashFunc(compilation);
                return true;
            }
        };
        /**
         * @internal
         */
        this.warnStandardHashFunc = (compilation) => {
            let foundStandardHashFunc = false;
            for (let i = 0; i < this.options.hashFuncNames.length; i += 1) {
                if (standardHashFuncNames.indexOf(this.options.hashFuncNames[i]) >= 0) {
                    foundStandardHashFunc = true;
                }
            }
            if (!foundStandardHashFunc) {
                this.warnOnce(compilation, "It is recommended that at least one hash function is part of the set " +
                    "for which support is mandated by the specification. " +
                    "These are: " +
                    standardHashFuncNames.join(", ") +
                    ". " +
                    "See http://www.w3.org/TR/SRI/#cryptographic-hash-functions for more information.");
            }
        };
        /**
         * @internal
         */
        this.validateHashFuncName = (compilation, hashFuncName) => {
            if (typeof hashFuncName !== "string" &&
                !(hashFuncName instanceof String)) {
                this.error(compilation, "options.hashFuncNames must be an array of hash function names, " +
                    "but contained " +
                    hashFuncName +
                    ".");
                return false;
            }
            try {
                crypto_1.createHash(hashFuncName);
            }
            catch (error) {
                this.error(compilation, "Cannot use hash function '" + hashFuncName + "': " + error.message);
                return false;
            }
            return true;
        };
        if (typeof options !== "object") {
            throw new Error("webpack-subresource-integrity: argument must be an object");
        }
        this.options = {
            hashFuncNames: ["sha384"],
            enabled: "auto",
            ...options,
        };
    }
    /**
     * @internal
     */
    emitMessage(messages, message) {
        messages.push(new Error(`${thisPluginName}: ${message}`));
    }
    /**
     * @internal
     */
    emitMessageOnce(messages, message) {
        if (!this.emittedMessages.has(message)) {
            this.emittedMessages.add(message);
            this.emitMessage(messages, message);
        }
    }
    /**
     * @internal
     */
    warnOnce(compilation, message) {
        this.emitMessageOnce(compilation.warnings, message);
    }
    /**
     * @internal
     */
    errorOnce(compilation, message) {
        this.emitMessageOnce(compilation.errors, message);
    }
    /**
     * @internal
     */
    error(compilation, message) {
        this.emitMessage(compilation.errors, message);
    }
    /**
     * @internal
     */
    warnIfHotUpdate(compilation, source) {
        if (source.indexOf("webpackHotUpdate") >= 0) {
            this.warnOnce(compilation, "webpack-subresource-integrity may interfere with hot reloading. " +
                "Consider disabling this plugin in development mode.");
        }
    }
    /**
     * @internal
     */
    updateAssetIntegrity(assetKey, integrity) {
        if (!this.assetIntegrity.has(assetKey)) {
            this.assetIntegrity.set(assetKey, integrity);
            this.inverseAssetIntegrity.set(integrity, assetKey);
        }
    }
    /**
     * @internal
     */
    isEnabled(compilation) {
        if (this.options.enabled === "auto") {
            return compilation.options.mode !== "development";
        }
        return this.options.enabled;
    }
    apply(compiler) {
        compiler.hooks.beforeCompile.tapPromise(thisPluginName, async () => {
            try {
                getHtmlWebpackPluginHooks = (await Promise.resolve().then(() => __importStar(require("html-webpack-plugin"))))
                    .default.getHooks;
            }
            catch (e) {
                if (e.code !== "MODULE_NOT_FOUND") {
                    throw e;
                }
            }
        });
        compiler.hooks.afterPlugins.tap(thisPluginName, (compiler) => {
            compiler.hooks.thisCompilation.tap({
                name: thisPluginName,
                stage: -10000,
            }, (compilation) => {
                this.setup(compilation);
            });
            compiler.hooks.compilation.tap("DefaultStatsFactoryPlugin", (compilation) => {
                compilation.hooks.statsFactory.tap(thisPluginName, (statsFactory) => {
                    statsFactory.hooks.extract
                        .for("asset")
                        .tap(thisPluginName, (object, asset) => {
                        if (this.assetIntegrity.has(asset.name)) {
                            object.integrity = String(this.assetIntegrity.get(asset.name));
                        }
                    });
                });
            });
        });
    }
}
exports.SubresourceIntegrityPlugin = SubresourceIntegrityPlugin;
//# sourceMappingURL=index.js.map